CREATE TABLE Gene (
  id int(11) NOT NULL auto_increment,
  ensembl_g_id varchar(100) NOT NULL,                                       #EnsEMBL gene ID
  ensembl_version varchar(100) NOT NULL,				    #Source EnsEMBL version
  source varchar(100) NOT NULL,						    #Source EnsEMBL database
  gene_type varchar(100) NOT NULL,					    #Gene type (protein_coding, pseudogene, miRNA, etc.)
  gene_name varchar(100) NOT NULL,					    #Gene symbol
  description text NOT NULL,                                                #Full description of gene - sometimes a long sentence
  evidence enum ('Known Gene','Unknown Gene', 'na') NOT NULL default 'na',  #Evidence.  Known or predicted gene according to EnsEMBL
  sequence longtext NOT NULL,						    #Complete gene sequence from first to last exon including introns
  chr_strand enum ('1','-1','na') NOT NULL default 'na',		    #Chromosome strand
  corrected_strand enum ('1','-1','na') NOT NULL default 'na',		    #All genes are stored in forward orientation.  If it had to changed this is noted here
  chromosome varchar(100) NOT NULL default 'na',				    #Chromosome name
  chr_start int(11) NOT NULL,						    #Start position of gene in chromosome coordinates 
  chr_end int(11) NOT NULL,						    #End position of gene in chromosome coordinates
  gene_start int(11) NOT NULL,						    #Start position in gene coordinates (always 1)
  gene_end int(11) NOT NULL,						    #End position in gene coordinates (start + length of all exons and introns)
  PRIMARY KEY (id),
  KEY ensembl_g_id (ensembl_g_id),
  KEY gene_name (gene_name) 
) ENGINE=InnoDB;

CREATE TABLE Transcript (
  id int(11) NOT NULL auto_increment,
  fk_Gene__id int(11) NOT NULL,				 #Foreign key to Gene table
  ensembl_t_id varchar(100) NOT NULL,			 #EnsEMBL transcript ID
  start int(11) NOT NULL,				 #Start position (in gene coordinates) of the transcript relative to the start of the gene (gene coordinates)
  end int(11) NOT NULL,					 #End position (in gene coordinates) of the transcript relative to the start of the gene (gene coordinates) 
  coding_region_start varchar(11) NOT NULL default 'na', #CDS start (in gene coordinates), will be 'na' for transcripts without a CDS
  coding_region_end varchar(11) NOT NULL default 'na',   #CDS end (in gene coordinates), will be 'na' for transcripts without a CDS
  cds_start_coords text NOT NULL,                        #Simple list of spliced start postions for CDS (in gene coordinates)
  cds_end_coords text NOT NULL,                          #Simple list of spliced end postions for CDS (in gene coordinates)
  PRIMARY KEY (id),
  KEY ensembl_t_id (ensembl_t_id),
  KEY fk_Gene__id (fk_Gene__id),
  CONSTRAINT FOREIGN KEY (`fk_Gene__id`) REFERENCES `Gene` (`id`)
) ENGINE=InnoDB;

CREATE TABLE Protein_feature (
  id int(11) NOT NULL auto_increment,
  fk_Transcript__id int(11) NOT NULL,  #Each feature is associated with a translation of a specific transcript (specified by this foreign key to the Transcript table)
  type varchar(100) NOT NULL,          #Known in EnsEMBL as 'logic_name' (e.g. Smart, Pfam, etc.)
  interpro_ac varchar(100) NOT NULL,   #Interpro accession, may be 'na' for some features
  name varchar(100) NOT NULL,          #Known in EnsEMBL as 'idesc' (e.g. Znf_C2H2), may be 'na' for some features
  program varchar(100) NOT NULL,       #program used to predict or annotate the feature (e.g. hmmpfam).  May be 'na' for some features
  start int(11) NOT NULL,              #Start position of the feature (in gene coordinates)
  end int(11) NOT NULL,                #End position of the feature (in gene coordinates)
  start_coords text NOT NULL,          #Simple list of spliced start postions (i.e. start and end coords define the exon content covered by this feature)
  end_coords text NOT NULL,            #Simple list of spliced end postions (i.e. start and end coords define the exon content covered by this feature)
  PRIMARY KEY (id),
  KEY fk_Transcript__id (fk_Transcript__id),
  CONSTRAINT FOREIGN KEY (`fk_Transcript__id`) REFERENCES `Transcript` (`id`)
) ENGINE=InnoDB;

CREATE TABLE External_id (
  id int(11) NOT NULL auto_increment,
  fk_Transcript__id int(11) NOT NULL,  #Foreign key to Transcript table
  type varchar(100) NOT NULL, 	       #Type of external ID (GO, Entrez, etc.)	
  external_id varchar(100) NOT NULL,   #ID from external database
  PRIMARY KEY (id),
  KEY fk_Transcript__id (fk_Transcript__id),
  CONSTRAINT FOREIGN KEY (`fk_Transcript__id`) REFERENCES `Transcript` (`id`)
) ENGINE=InnoDB;

CREATE TABLE Exon (
  id int(11) NOT NULL auto_increment,
  ensembl_e_id varchar(100) NOT NULL,  #EnsEMBL exon ID
  start int(11) NOT NULL,              #Start position of exon (in gene coordinates)
  end int(11) NOT NULL,                #End position of exon (in gene coordinates)
  PRIMARY KEY (id),
  KEY ensembl_e_id (ensembl_e_id)
) ENGINE=InnoDB;

CREATE TABLE TranscriptExon (
  id int(11) NOT NULL auto_increment,  #Defines a connection between transcripts and exons (many-to-many relationships are possible)
  fk_Transcript__id int(11) NOT NULL,  #Foreign key to Transcript table
  fk_Exon__id int(11) NOT NULL,        #Foreign key to Exon table
  PRIMARY KEY (id),
  KEY fk_Transcript__id (fk_Transcript__id),
  KEY fk_Exon__id (fk_Exon__id),
  CONSTRAINT FOREIGN KEY (`fk_Transcript__id`) REFERENCES `Transcript` (`id`),
  CONSTRAINT FOREIGN KEY (`fk_Exon__id`) REFERENCES `Exon` (`id`)
) ENGINE=InnoDB;

#Please note the following for the 'Probe' table
#1.) 'probe_count' is the probe ID taken from the probe file used to import a batch of probes.  It would ideally be the same as the auto (id) but just 
#    in case, this number is stored to ensure that I can correlate probe info in the database back to the source probe files
#2.) Probe parameters are determined by a number of different tools and algorithms
#3.) Tm values are determined by the nearest neighbour method - see Probes.pm for details
#    RNA folding characteristics are determined in two ways - one looks for internal probe folding and the other looks for self-self
#    pairing between two probe molecules.  See get_SimFold-PairFoldScores.pl for details.  These scores are generally a negative number
#    the more negative the number the greater the strength of the fold and therefore the less desirable the probe
#4.) Probes that correspond to Repeat regions of the genome are filtered out by the scripts that select probe sequences by comparing to repeatmasked 
#    gene sequences.  A small number of these 'masked_bases' are allowed (up to 1/4 of the length of the probe) 
#5.) Simple repeats that might be missed by the masking (homopolymers, dinucleotide repeats, etc.) are detected using 
#    the program 'mdust'.  The mdust score represents the number of bases masked by mdust as low complexity at the specified mdust threshold 
#6.) The ProbeSet and Set tables will be used to define arbitrary groups of probes (other than the probes of a microarray design)
#7.) 'target' hits are blast similarity hits to the gene sequence the probe is supposed to detect!
#    'non-target' hits are blast similarity hits to another sequence (an EST from a different gene for example)
#    Since negative control probes have no target gene, all of their hits are 'non-target' hits.
CREATE TABLE Probe (
  id int(11) NOT NULL auto_increment,
  probe_count int(11) NOT NULL default '0',	   #Probe ID from source probe file
  probe_set_id int(11) NOT NULL default '0', 	   #ProbeSet ID from source probe file (corresponds to a single exon, intron, junction or boundary) 
  sequence text NOT NULL,                          #Probe sequence
  probe_length int(11) NOT NULL default '0',       #Probe length
  tm float NOT NULL,                               #Tm of probe in degress C 
  type enum ('Exon','Exon-Exon','Exon-Intron','Intron-Exon','Intron','Control-Negative','na') NOT NULL default 'na',
  masked_bases int(11) NOT NULL,                   #Number of bases of this probe that were masked in a masked version of the source gene sequence
  sim_fold_score float NOT NULL,                   #Hairpin folding score from SimFold in kcal/mol
  pair_fold_score float NOT NULL,                  #Dimerization fold potential from PairFold in kcal/mol
  mdust_bases int(11) NOT NULL,			   #Number of low complexity bases in the probe sequence according to 'mdust'
  mrna_target_hit_length int(11),                  #Longest blast hit observed for this probe to mRNAs mapping to the TARGET gene 
  mrna_non_target_hit_length int(11),		   #Longest blast hit observed for this probe to mRNAs mapping to genes OTHER THAN the target 
  est_target_hit_length int(11),		   #Longest blast hit observed for this probe to ESTs mapping to the TARGET gene 
  est_non_target_hit_length int(11),		   #Longest blast hit observed for this probe to ESTs mapping to genes OTHER THAN the target 
  enst_target_hit_length int(11),		   #Longest blast hit observed for this probe to EnsEMBL transcripts mapping to the TARGET gene 
  enst_non_target_hit_length int(11),		   #Longest blast hit observed for this probe to EnsEMBL transcripts mapping to genes OTHER THAN the target 
  probe_target_hit_length int(11),                 #Longest blast hit observed for this probe to its OWN ProbeSet 
  probe_non_target_hit_length int(11),		   #Longest blast hit observed for this probe to its ANOTHER ProbeSet 
  genomic_non_target_hit_length int(11), 	   #Longest blast hit observed for this probe to the complete genome (only used for random negative control probes)
  within_gene_target_hit_length int(11),           #Longest blast hit observed for this probe to the region of its complete target gene sequence it was extracted from
  within_gene_non_target_hit_length int(11),       #Longest blast hit observed for this probe to any region of its complete target gene sequence OTHER THAN where it was extracted from (e.g. another exon for example)
  exons_skipped varchar(11) NOT NULL default 'na', #Number of exons skipped in the hypothetical junction this probe represents (exon-junction probes only) 
  PRIMARY KEY (id),
  KEY probe_count (probe_count)
) ENGINE=InnoDB;

#The set table is used to define Sets of probes.
#These could be sets of probes defined by the analysis, sets of DE probes, sets of probes passing some complicated test, etc.
#Note: The set of probes defining a particular microarray design will be defined by the ArraySpot and Microarray tables
CREATE TABLE Probe_set (
  id int(11) NOT NULL auto_increment,  
  type text NOT NULL,                   #Type of 'Probe_set' being defined.  All probes which pass filter tests are a set for example
  description text NOT NULL,            #Description of this set.  For example, a description of the filtering parameters used to define the set
  PRIMARY KEY (id)
) ENGINE=InnoDB;

#Allows many to many relationships between probes and probe sets
CREATE TABLE ProbeProbe_set (
  id int(11) NOT NULL auto_increment,
  fk_Probe_set__id int(11) NOT NULL,   #Foreign key to Probe_set table
  fk_Probe__id int(11) NOT NULL,       #Foreign key to Probe table
  PRIMARY KEY (id),
  KEY fk_Probe_set__id (fk_Probe_set__id),
  KEY fk_Probe__id (fk_Probe__id),
  CONSTRAINT FOREIGN KEY (`fk_Probe_set__id`) REFERENCES `Probe_set` (`id`),
  CONSTRAINT FOREIGN KEY (`fk_Probe__id`) REFERENCES `Probe` (`id`)
) ENGINE=InnoDB;

#Note: Probes will be linked to Genes not to individual exons because this can lead to ambiguous relationships
#This is because ensembl contains exons which have identical sequence from the same gene but are associated with 
#different transcripts and are therefore given different exon IDs because their protein coding context is different
#If the protein coding context is the same for two exons in different transcripts then the same ID is used, but this is
#often not the case.  To determine the relevant exons for a probe will thus require considering the sequence position 
#within the specified gene.
#Junction probes are made up of two units (exon1-exon2 or exon-intron) - A unit can thus refer to a piece of exon or intron sequence
#Non-junction probes such as exon and intron probes consist of only a single unit, unit2 values will be 'na'
#Negative control probes do not match the genome or any specific gene so they are not referenced in this table
CREATE TABLE GeneProbe (
  id int(11) NOT NULL auto_increment,
  fk_Gene__id int(11) NOT NULL,       #Foreign key to Gene table
  fk_Probe__id int(11) NOT NULL,      #Foreign key to Probe table
  unit1_start text NOT NULL,          #Start position of UNIT1 of probe within gene (to help identify exon being targetted)   
  unit1_end text NOT NULL,            #End position of UNIT1 of probe within gene (to help identify exon being targetted)
  unit1_start_chr text NOT NULL,      #Start position of UNIT1 of probe within chromosome (for direct mapping to genome)   
  unit1_end_chr text NOT NULL,        #End position of UNIT1 of probe within chromosome (for direct mapping to genome)
  exon1_ids text NOT NULL,            #Exon IDs corresponding to targeted sequence, multiple exons occur for exons with the same genomic END position
  unit2_start text NOT NULL,          #Start position of UNIT2 of probe within gene (to help identify exon or intron being targetted - may be 'na')   
  unit2_end text NOT NULL,            #End position of UNIT2 of probe within gene (to help identify exon or intron being targetted - may be 'na')	
  unit2_start_chr text NOT NULL,      #Start position of UNIT2 of probe within chromosome (for direct mapping to genome)   
  unit2_end_chr text NOT NULL,        #End position of UNIT2 of probe within chromosome (for direct mapping to genome)
  exon2_ids text NOT NULL,            #Exon IDs corresponding to targeted sequence, multiple exons occur for exons with the same genomic START position  
  PRIMARY KEY (id),
  KEY fk_Gene__id (fk_Gene__id),
  KEY fk_Probe__id (fk_Probe__id),
  CONSTRAINT FOREIGN KEY (`fk_Gene__id`) REFERENCES `Gene` (`id`),
  CONSTRAINT FOREIGN KEY (`fk_Probe__id`) REFERENCES `Probe` (`id`)
) ENGINE=InnoDB;

CREATE TABLE Microarray (
  id int(11) NOT NULL auto_increment,
  name varchar(100) NOT NULL,
  type varchar(100) NOT NULL,
  manufacturer varchar(100) NOT NULL,
  probe_length varchar(100) NOT NULL,
  order_date date NOT NULL,  
  PRIMARY KEY (id)
) ENGINE=InnoDB;

CREATE TABLE ArraySpot (
  id int(11) NOT NULL auto_increment,
  fk_Microarray__id int(11) NOT NULL,
  fk_Probe__id int(11) NOT NULL,
  type enum ('PROBE','EMPTY','na') NOT NULL default 'na',
  x_coord int(11) NOT NULL,
  y_coord int(11) NOT NULL,
  array_row int(11) NOT NULL,
  array_column int(11) NOT NULL,
  PRIMARY KEY (id),
  KEY fk_Microarray__id (fk_Microarray__id),
  KEY fk_Probe__id (fk_Probe__id),
  CONSTRAINT FOREIGN KEY (`fk_Microarray__id`) REFERENCES `Microarray` (`id`),
  CONSTRAINT FOREIGN KEY (`fk_Probe__id`) REFERENCES `Probe` (`id`)
) ENGINE=InnoDB;

CREATE TABLE MaskedGene (
  id int(11) NOT NULL auto_increment,
  fk_Gene__id int(11) NOT NULL,
  sequence longtext NOT NULL,
  PRIMARY KEY (id),
  KEY fk_Gene__id (fk_Gene__id),
  CONSTRAINT FOREIGN KEY (`fk_Gene__id`) REFERENCES `Gene` (`id`)
) ENGINE=InnoDB;

#Table downloaded directly from UCSC containing alignments of all human mRNAs to the human genome (Build35/hg17)
CREATE TABLE all_mrna (
  bin smallint(5) unsigned NOT NULL default '0',
  matches int(10) unsigned NOT NULL default '0',
  misMatches int(10) unsigned NOT NULL default '0',
  repMatches int(10) unsigned NOT NULL default '0',
  nCount int(10) unsigned NOT NULL default '0',
  qNumInsert int(10) unsigned NOT NULL default '0',
  qBaseInsert int(10) unsigned NOT NULL default '0',
  tNumInsert int(10) unsigned NOT NULL default '0',
  tBaseInsert int(10) unsigned NOT NULL default '0',
  strand char(2) NOT NULL default '',
  qName varchar(255) NOT NULL default '',
  qSize int(10) unsigned NOT NULL default '0',
  qStart int(10) unsigned NOT NULL default '0',
  qEnd int(10) unsigned NOT NULL default '0',
  tName varchar(255) NOT NULL default '',
  tSize int(10) unsigned NOT NULL default '0',
  tStart int(10) unsigned NOT NULL default '0',
  tEnd int(10) unsigned NOT NULL default '0',
  blockCount int(10) unsigned NOT NULL default '0',
  blockSizes longblob NOT NULL,
  qStarts longblob NOT NULL,
  tStarts longblob NOT NULL,
  KEY tName (tName(8),bin),
  KEY tName_2 (tName(8),tStart),
  KEY qName (qName(12)),
  KEY tName_3 (tName(8),tEnd)
) ENGINE=MyISAM;

#Table downloaded directly from UCSC containing alignments of all human ESTs to the human genome (Build35/hg17)
CREATE TABLE all_est (
  bin smallint(5) unsigned NOT NULL default '0',
  matches int(10) unsigned NOT NULL default '0',
  misMatches int(10) unsigned NOT NULL default '0',
  repMatches int(10) unsigned NOT NULL default '0',
  nCount int(10) unsigned NOT NULL default '0',
  qNumInsert int(10) unsigned NOT NULL default '0',
  qBaseInsert int(10) unsigned NOT NULL default '0',
  tNumInsert int(10) unsigned NOT NULL default '0',
  tBaseInsert int(10) unsigned NOT NULL default '0',
  strand char(2) NOT NULL default '',
  qName varchar(255) NOT NULL default '',
  qSize int(10) unsigned NOT NULL default '0',
  qStart int(10) unsigned NOT NULL default '0',
  qEnd int(10) unsigned NOT NULL default '0',
  tName varchar(255) NOT NULL default '',
  tSize int(10) unsigned NOT NULL default '0',
  tStart int(10) unsigned NOT NULL default '0',
  tEnd int(10) unsigned NOT NULL default '0',
  blockCount int(10) unsigned NOT NULL default '0',
  blockSizes longblob NOT NULL,
  qStarts longblob NOT NULL,
  tStarts longblob NOT NULL,
  KEY tName (tName(8),bin),
  KEY tName_2 (tName(8),tStart),
  KEY qName (qName(12)),
  KEY tName_3 (tName(8),tEnd)
) ENGINE=MyISAM;

#DROP TABLE MaskedGene,ArraySpot,Microarray,GeneProbe,ProbeProbe_set,Probe,Probe_negativeControl,Probe_set,TranscriptExon,Exon,Transcript,External_id,Gene;
#DROP TABLE ArraySpot,Microarray,GeneProbe,ProbeProbe_set,Probe,Probe_set
